#!/usr/bin/env python
# Copyright C 2015 by Isilon Systems, Inc. All rights not expressly
# granted herein are reserved.
# Created in the United States of America.
#
# This digital media is protected by U.S. and international copyright and
# intellectual property laws. Unless otherwise specified, all information
# and screens appearing as part of this digital medium, including
# software, services, documents, text, images, icons, and logos design;
# the selection, assembly, arrangement, and design thereof; and the code
# that enables its presentation, are the sole property of Isilon Systems,
# Inc.
#
# Except as otherwise required by applicable law, no part of any
# copyrighted material may be reproduced, modified, translated,
# distributed, published, or transmitted by any means, electronic or
# mechanical, for any purpose, without the prior written consent of Isilon
# Systems.
#
# Isilon and Isilon Systems are registered trademarks of Isilon Systems,
# Inc.  "Isilon IQ" is a trademark of Isilon Systems, Inc.  All other
# trademarks and registered trademarks are the property of their
# respective holders.
"""
Read sparse file(s) with dd and measure performance
"""
__author__ = 'luthj'

import datetime
import logging
import multiprocessing as mp
import optparse  # argparse not available until Python 2.7
import os
import re
import subprocess
import sys
import tempfile
import time

# Python <= 2.5 uses Simple JavaScript Object Notation (simplejson). A newer
# simplejson called json is used from Python 2.6
try:
    import json
except ImportError:
    import simplejson as json

DD_READ_SPARSE_RESULTS_FILE = "dd_read_sparse_results.csv"
DD_READ_SPARSE_DIR = "dd_files"
FILE_DIR = '/tmp'
DESIRED_RUNTIME_S = 10
# MB/s throughput assumption to adjust parameters to desired runtime
# Note that this is based on large (1M) blocksize
ASSUMED_THROUGHPUT_MBPS = 4000
ASSUMED_MAX_THROUGHPUT_MBPS = 20000

# This needs to be outside of the class so it can be passed to
# multiprocess.pool.map()
def read_file(params):
    """
    Read from a sparse file using dd
    A dictionary is used for parameters because multiprocessor.pool()
    requires an iterable for command arguments
    :param params: A dictionary of the dd params
    :return: dd output string
    """
    # TODO: better validation of params/error handling
    error = 0
    if 'filepath' in params:
        filepath = params['filepath']
    else:
        sys.stderr.write('file path not found')
        error = 1
    if 'blocksize' in params:
        blocksize = params['blocksize']
    else:
        sys.stderr.write('blocksize not found')
        error = 1
    if 'count' in params:
        count = params['count']
    else:
        sys.stderr.write('count not found')
        error = 1
    if error:
        exit(2)
    cmd = ['time', 'dd','if=' + filepath, 'of=/dev/null',
           'bs=' + str(blocksize), 'count=' + str(count)]
    p1 = subprocess.Popen(cmd,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)
    out, err = p1.communicate()
    return err

class PfOptionParser(optparse.OptionParser):
    """An option parser that supports post-parse validation hooks.
    Validators must be callable objects accepting two arguments: an
    OptionValues object and a list of arguments.  They may modify
    their inputs.  To signal an error, they should raise
    optparse.OptionValueError.
    """

    def __init__(self, **kwargs):
        if 'formatter' not in kwargs:
            kwargs['formatter'] = BetterHelpFormatter()
        optparse.OptionParser.__init__(self, **kwargs)
        self.__validators = []

    def append_validator(self, validator):
        self.__validators.append(validator)

    def check_values(self, options, args):
        # This is called by optparse.OptionParser.parse_args.
        try:
            for validator in self.__validators:
                validator(options, args)
        except (optparse.BadOptionError, optparse.OptionValueError), err:
            self.error(str(err))
        return (options, args)

class BetterHelpFormatter(optparse.IndentedHelpFormatter):
    def __init__(self, *args, **kwargs):
        optparse.IndentedHelpFormatter.__init__(self, *args, **kwargs)

    def format_epilog(self, epilog):
        if epilog:
            paragraphs = epilog.split("\n\n")
            epilog = ""
            for p in paragraphs:
                epilog += "\n"
                epilog += optparse.IndentedHelpFormatter._format_text(self, p)
                epilog += "\n"
        else:
            epilog = ""
        return epilog

class DdReadSparse(object):
    """
    Python wrapper for running parallel dd's against sparse files
    """

    def __init__(self, args=None):
        """
        Constructor
        """
        self.__dd_procs = None
        self.__blocksize_B = None
        self.__dd_count = None
        self.__filesize_MB = None
        self.created_files = list()
        self.total_runtime = None

        self.extra_files = []
        self.temp_path = tempfile.gettempdir()
        self.options, self.args = self.get_options(args)

    @property
    def dd_procs(self):
        """
        Get the dd_procs
        """
        return self.__dd_procs

    @dd_procs.setter
    def dd_procs(self, value):
        """
        Set the dd_procs
        """
        self.__dd_procs = int(value)

    @property
    def blocksize_B(self):
        """
        Get the blocksize_B
        """
        return self.__blocksize_B

    @blocksize_B.setter
    def blocksize_B(self, value):
        """
        Set the blocksize_B
        """
        self.__blocksize_B = int(value)

    @property
    def dd_count(self):
        """
        Get the dd_count
        """
        return self.__dd_count

    @dd_count.setter
    def dd_count(self, value):
        """
        Set the dd_count
        """
        self.__dd_count = int(value)

    @property
    def filesize_MB(self):
        """
        Get the filesize_MB
        """
        return self.__filesize_MB

    @filesize_MB.setter
    def filesize_MB(self, value):
        """
        Set the filesize_MB
        """
        self.__filesize_MB = int(value)

    def get_options(self, args=None):
        if args is None:
            args = sys.argv[1:]
        return self.get_parser().parse_args(args)

    def get_parser(self, *parser_args, **parser_kwargs):
        """
        Parse args
        """
        parser = PfOptionParser(*parser_args, **parser_kwargs)
        parser.append_validator(self.validate_options)
        # TODO: make count a parameter?
        parser.add_option(
            "--resultsdir", dest="resultsdir", default=None,
            help="directory in which to store test results")
        parser.add_option(
            "--dd_procs", dest="dd_procs", default=None,
            help="simultaneous dd processes to run")
        dd_group = optparse.OptionGroup(
            parser, "dd options")
        dd_group.add_option(
            "--blocksize", dest="blocksize", default="1M",
            help="dd blocksize")
        parser.add_option_group(dd_group)
        misc_group = optparse.OptionGroup(parser, "miscellaneous options")
        misc_group.add_option(
            "-E", "--extra", dest="extra", default=None,
            help="location to store results (error messages, I/O timestamps)")
        parser.add_option_group(misc_group)
        return parser

    def validate_options(self, options, args):
        """
        Validate args
        """
        if not options.dd_procs:
            raise optparse.OptionValueError('--dd_procs is required')
        try:
            int(options.dd_procs)
            if options.dd_procs < 1:
                raise optparse.OptionValueError('--dd_procs must be > 1')
        except ValueError:
            raise optparse.OptionValueError('--dd_procs must be an integer')
        #TODO: is 4 * ncpu an appropriate procs limit?
        if int(options.dd_procs) > (4 * self.count_cores()):
            raise optparse.OptionValueError('--dd_procs must <= 4 * ncpu')
        self.dd_procs = options.dd_procs
        self.blocksize_B = self.human2number(options.blocksize)
        self.dd_count = self.calculate_dd_count()

    def flush_node_cache(self):
        """Flush any of the cluster's caches required before testing.

        This implementation flushes L2 twice and then L3.

        If metadata need to be populated as part of the flushing process,
        it needs to be done in this method override.
        """
        wrapper.BaseWrapper.flush_node_cache(self)

    def flush_client_cache(self):
        """Flush any of the client's caches required before testing.

        This implementation flushes the client caches for the partition
        containing the target filename.
        """
        wrapper.BaseWrapper.flush_client_cache(self)

    @property
    def util_name(self):
        """
        Return the name of the test program
        """
        return "dd"

    @property
    def util_cmd(self):
        """The utility command without any additional arguments.
        If --location was not passed to the wrapper then the utility
        should be on the system path.
        """
        # This checks for the existence of self.options because it
        # is called early in initialization.
        if hasattr(self, 'options') and \
                hasattr(self, 'options.location') and \
                        self.options.location is not None:
            return self.options.location
        return self.util_name + (".exe" if pfutils.on_windows() else "")

    def get_util_cmd_args(self):
        """
        We don't use this, just need to satisfy the abstract method
        """
        o = self.options
        args = []

        return args

    def parse_run_results(self, results):
        """
        Parse the output from the parallel dd runs
        """
        parsed = list()
        for result in results:
            parsed.append(self.parse_dd_output(result))
        return parsed

    def prepare_test(self):
        """
        Set up the results and target directories
        """
        if self.options.extra:
            for base in [self.options.extra, self.temp_path]:
                path = os.path.join(base, DD_READ_SPARSE_DIR)
                if not os.path.exists(path):
                    os.makedirs(path)
                    logging.debug("Created directory %s.", path)
        if self.options.resultsdir:
            if not os.path.exists(self.options.resultsdir):
                os.makedirs(self.options.resultsdir)
                logging.debug("Created directory %s.",
                              self.options.resultsdir)
        else:
                logging.debug("resultsdir not supplied, \
                    skipping creation of resultsdir")

        # Create the directory for the sparse files
        if not os.path.exists(FILE_DIR):
            os.makedirs(FILE_DIR)

        # Calculate the sparse file size
        self.filesize_MB = self.calculate_file_size_mb()
        return

    def cleanup_test(self):
        """
        Clean up after the run
        :return:
        """
        self.delete_created_files()

    # Like format_run_results(), but for pertinent test run parameters
    def format_run_options(self):
        """
        Put the test options in a JSON format
        Like format_run_results(), but for pertinent test run parameters
        """
        o_dict = {"procs" : self.dd_procs,
                  "blocksize" : self.blocksize_B,
                  "count"   : self.dd_count}
        return json.dumps(o_dict)

    def format_run_results(self):
        """
        Combine the results from all the parallel dd runs into aggregate
        :return: JSON dump of the aggregate results
        """
        combined = dict()
        combined['bytes'] = 0
        for result in self.run_results:
            combined['bytes'] += result['bytes']
        combined['total_runtime'] = self.total_runtime
        combined['total_MBps'] = combined['bytes']/1024/1024/self.total_runtime
        combined['per_proc_MBps'] = combined['total_MBps']/self.dd_procs

        return json.dumps(combined)

    def log_results(self):
        """
        Store the test options and results in a CSV file
        This makes importing the results into a spreadsheet easier
        """
        # TODO: store the results of each individual dd run instead of agg?
        if not self.options.resultsdir:
            return
        logfile = self.options.resultsdir + "/" + \
                  DD_READ_SPARSE_RESULTS_FILE
        results_csv = self.json_to_csv(self.format_run_results())
        options_csv = self.json_to_csv(self.format_run_options())
        # If the results file doesn't exist, print the column headers first
        if not os.path.isfile(logfile):
            with open(logfile, "a") as results_file:
                results_file.write(results_csv[0] + ",")
                results_file.write(options_csv[0] + "\n")
        with open(logfile, "a") as results_file:
            results_file.write(results_csv[1] + ",")
            results_file.write(options_csv[1] + "\n")
        return

    def save_logs(self):
        """
        Save results data
        """
        self.log_results()
        return

    # TODO: belongs in a library
    def json_to_csv(self, json_string):
        """
        Convert JSON data to CSV format
        :param json_string: a string in JSON format
        :return: string in CSV format
        """
        my_dict = json.loads(json_string)
        columns = list(my_dict.keys())
        csv_output = ["", ""]
        for col in columns:
            csv_output[0] += str(col) + ','
        for col in columns:
            # FIXME: Ending up with a trailing '%' parsing the options list,
            # not sure why
            csv_output[1] += str(my_dict[col]) + ','
        return csv_output

    # TODO: belongs in a library
    def count_cores(self):
        """
        Get the number of cores on the system (sysctl hw.ncpu)
        :return: number of cores
        """
        num_cores = None
        p1 = subprocess.Popen(['sysctl', 'hw.ncpu'],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        out, err = p1.communicate()
        if out.strip() != '':
            num_cores = out.strip().split(': ')[1]
            try:
                num_cores = int(num_cores)
            except ValueError:
                num_cores = None
        return num_cores

    # TODO: belongs in a library
    def get_physmem(self):
        """
        Get the physical memory on the system (sysctl hw.ncpu)
        :return: physical memory in Bytes
        """
        memory = None
        p1 = subprocess.Popen(['sysctl', 'hw.physmem'],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        out, err = p1.communicate()
        if out.strip() != '':
            memory = out.strip().split(': ')[1]
            try:
                memory = long(memory)
            except ValueError:
                memory = None
        return memory

    # TODO: belongs in a library
    # and its inverse
    def human2number(self, s):
        """
        Converts a number with a prefix to units ('k', 'M', etc) to a number
        :param s: string representation of a number
        :return: a float value of the number
        """
        try:
            return int(s)
        except:
            pass
        SYMBOLS = {
            'customary'     : ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'),
            'customary_ext' : ('byte', 'kilo', 'mega', 'giga', 'tera',
                               'peta', 'exa','zetta', 'iotta'),
            'iec'           : ('Bi', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei',
                               'Zi', 'Yi'),
            'iec_ext'       : ('byte', 'kibi', 'mebi', 'gibi', 'tebi',
                               'pebi', 'exbi', 'zebi', 'yobi'),
        }
        init = s
        num = ''
        while s and s[0:1].isdigit() or s[0:1] == '.':
            num += s[0]
            s = s[1:]
        num = float(num)
        letter = s.strip()
        for name, sset in SYMBOLS.items():
            if letter in sset:
                break
        else:
            if letter == 'k':
                # treat 'k' as an alias for 'K' as per: http://goo.gl/kTQMs
                sset = SYMBOLS['customary']
                letter = letter.upper()
            else:
                raise ValueError("can't interpret %r" % init)
        prefix = {sset[0]:1}
        for i, s in enumerate(sset[1:]):
            prefix[s] = 1 << (i+1)*10
        return float(num * prefix[letter])

    def calculate_file_size_mb(self):
        """
        Calculate the appropriate file size to use for the test
        One file/proc
        :return: size of each file in MBytes
        """
        return (self.calculate_dd_count() * self.blocksize_B) / 1024 / 1024 

    def calculate_dd_count(self):
        """
        Attempt to tune dd count param to get a DESIRED_RUNTIME_S runtime
        Using procs**2.5 was determined through experimentation
        :return: value to be used for dd count parameter
        """
        # TODO - this could be smarter
        # (needs a different calc when procs > ncpus)
	throughput = ASSUMED_THROUGHPUT_MBPS * self.dd_procs
	if (throughput > ASSUMED_MAX_THROUGHPUT_MBPS):
		throughput = ASSUMED_MAX_THROUGHPUT_MBPS
	throughput = throughput / self.dd_procs
	count = (throughput * 1024 * 1024  * DESIRED_RUNTIME_S) / self.blocksize_B
	
        return int(count)

    def create_sparse_file(self, filenum):
        """
        Create a sparse file for the test
        :param filenum: file number index (one file per proc)
        :return: Full path of file
        """
        filepath = FILE_DIR + '/sparse_' + str(filenum)
        # Remove any prior file
        try:
            os.remove(filepath)
        except OSError:
            pass
        # Create the sparse file with dd
        size_1k_blks = self.filesize_MB*1024
        cmd = ['time','dd','if=/dev/zero', 'of=' + filepath, 'count=0',
               'bs=1k', 'seek=' + str(size_1k_blks)]
        p1 = subprocess.Popen(cmd,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        out, err = p1.communicate()
        self.created_files.append(filepath)
        return filepath

    def delete_created_files(self):
        """
        Delete all of the sparse files created for this test
        """
        for filepath in self.created_files:
            if os.path.dirname(filepath) == FILE_DIR:
                try:
                    os.remove(filepath)
                except OSError:
                    pass
        return

    def parallel_dd_reads(self):
        """
        Create sparse files (one/proc) and
        run dd reads in parallel using multiprocessor.pool()
        :return: a list of parsed dd results
        """
        params = list()
        for i in range(self.dd_procs):
            proc = i + 1
            param = dict()
            param['filepath'] = self.create_sparse_file(proc)
            param['blocksize'] = self.blocksize_B
            #param['count'] = int(COUNT_PER_PROC/self.dd_procs)
            param['count'] = self.dd_count
            params.append(param)
        pool = mp.Pool()
        start = time.time()
        results = pool.map(read_file, params)
        self.total_runtime = time.time() - start

        return self.parse_run_results(results)

    def parse_dd_output(self, output):
        """
        Parse dd output
        :param output: dd output to parse
        :return: dictionary of dd results
        """
        results = dict()
        p_throughput = re.compile(r'^(\d+)\sbytes\stransferred\sin\s(\d+\.?\d*)\ssecs\s+\((\d+)\sbytes\/sec.*$')
        for out_line in output.split("\n"):
            if p_throughput.match(out_line):
                results['bytes'] = long(p_throughput.match(out_line).group(1))
                results['seconds'] = float(
                    p_throughput.match(out_line).group(2))
                results['Bps'] = long(p_throughput.match(out_line).group(3))
        return results

    def run(self):
        """
        Run the test
        """
        #self.mod_early()
        self.prepare_test()
        #if self.options.flush:
        #    self.flush_node_cache()
        #else:
        #    self.pre_cache()
        #self.start_perf_collector()
        #self.mod_pre_test()
        #self.pre_test()
        #self.start_test()
        #self.communicate()
        #self.run_results = self.finish_test()
        self.run_results = self.parallel_dd_reads()
        self.timestamp = datetime.datetime.now().isoformat()
        #self.post_test()
        #self.mod_post_test()
        #self.stop_perf_collector()
        self.save_logs()
        self.cleanup_test()
        #self.mod_final()
        print "%s" % (self.format_run_results())
        #self.print_run_results()

if __name__ == '__main__':
    DdReadSparse().run()
